(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * Stateful.thy ---
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2010 ETH Zurich, Switzerland
 *               2008-2010 Achim D. Brucker, Germany
 *               2009-2010 Univerisite Paris-Sud, France 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: Stateful.thy 8835 2010-03-15 22:48:23Z brucker $ *)

theory Stateful 
imports "../PacketFilter/PacketFilter" Testing
begin

text{* The simple system of a stateless packet filter is not enough to
  model all common real-world scenarios. Some protocols need further
  actions in order to be secured.  A prominent example is the File
  Transfer Protocol (FTP), which is a popular means to move files
  across the Internet. It behaves quite differently from most other
  application layer protocols as it uses a two-way connection
  establishment which opens a dynamic port. A stateless packet filter
  would only have the possibility to either always open all the
  possible dynamic ports or not to allow that protocol at all. Neither
  of these options is satisfactory. In the first case, all ports above
  1024 would have to be opened which introduces a big security hole in
  the system, in the second case users wouldn't be very happy. A
  firewall which tracks the state of the TCP connections on a system
  doesn't help here either, as the opening and closing of the ports
  takes place on the application layer. Therefore, a firewall needs to
  have some knowledge of the application protocols being run and track
  the states of these protocols. We next model this behaviour.

  The key point of our model is the idea that a policy remains the
  same as before: a mapping from packet to packet out. We
  still specify for every packet, based on its source and
  destination address, the expected action.  The only thing that changes now
  is that this mapping is allowed to change over time. This indicates
  that our test data will not consist of single packets but rather of
  sequences thereof. 
*}

text{* 
  At first we hence need a state. It is a tupel from some memory to be
  refined later and the current policy. 
*}


types ('\<alpha>,'\<beta>,'\<gamma>) FWState = "'\<alpha> \<times> ('\<beta>,'\<gamma>) Policy"  



text{* Having a state, we need of course some state transitions. Such
  a transition can happen every time a new packet arrives. State
  transitions can be modelled using a state-exception monad.  *}

types ('\<alpha>,'\<beta>,'\<gamma>) FWStateTransition = "('\<beta>,'\<gamma>) packet \<Rightarrow> (unit, ('\<alpha>,'\<beta>,'\<gamma>) FWState) MON_SE"

text{* 
  The memory could be modelled as a list of accepted packets.
*}
types ('\<beta>,'\<gamma>) history = "('\<beta>,'\<gamma>) packet list"


text{* 
  The next two constants will help us later in defining the state
  transitions.  The constant @{text "before"} is @{text "True"} if for
  all elements which appear before the first element for which 
  @{text q} holds, @{text p} must hold.
*}

consts before :: "('\<alpha> \<Rightarrow> bool) \<Rightarrow> ('\<alpha> \<Rightarrow> bool) \<Rightarrow> '\<alpha> list \<Rightarrow> bool"
primrec 
  "before p q [] = False"
  "before p q (a # S) = (q a \<or> (p a \<and> (before p q S)))"

text{* 
  Analogously there is an operator @{text not_before} which returns
  @{text "True"} if for all elements which appear before the first
  element for which @{text q} holds, @{text p} must not hold.
*}
consts not_before :: "('\<alpha> \<Rightarrow> bool) \<Rightarrow> ('\<alpha> \<Rightarrow> bool) \<Rightarrow> '\<alpha> list \<Rightarrow> bool" 
primrec 
  "not_before p q [] = False"
  "not_before p q (a # S) = (q a \<or> (\<not> (p a) \<and> (not_before p q S)))"

text{* 

  The next two operators can be used to combine state transitions. It
  takes the first transition which maps to @{text "Some '\<alpha>"}. *}

definition orelse:: "('\<alpha>,'\<beta>,'\<gamma>) FWStateTransition \<Rightarrow> ('\<alpha>,'\<beta>,'\<gamma>) FWStateTransition \<Rightarrow> ('\<alpha>,'\<beta>,'\<gamma>) FWStateTransition"  (infixl "orelse" 100) where
"(f orelse g) x \<equiv> \<lambda> \<sigma>. (case f x \<sigma> of None \<Rightarrow> g x \<sigma> | Some y \<Rightarrow> Some y)"

end
